use std::{
    fs::File,
    io::{BufRead, BufReader},
    sync::{Arc, Mutex},
    thread,
};
use std::collections::HashMap;
use std::time::Instant;

const DEFAULT_HASHMAP_LENGTH: usize = 10000;

fn main() {
    let now = Instant::now();
    let stations: Arc<Mutex<HashMap<String, onebrc::StationMeasurements>>> =
        Arc::new(Mutex::new(HashMap::with_capacity(DEFAULT_HASHMAP_LENGTH)));

    let cores: usize = thread::available_parallelism().unwrap().into();

    let chunk_length = 1_000_000_000 / cores;
    let mut handles = vec![];
    for i in 0..cores {
        let file = File::open("../../../measurements.txt").expect("File measurements.txt not found");
        let reader = BufReader::new(file);
        let line_chunk = reader.lines().skip(chunk_length * i).take(chunk_length);
        let stations_clone = stations.clone();
        let handle = thread::spawn(move || {
            let mut t_stations: HashMap<String, onebrc::StationMeasurements> =
                HashMap::with_capacity(DEFAULT_HASHMAP_LENGTH);

            let now_read_line = Instant::now();
            println!("Start reading lines in thread {i}");
            line_chunk.for_each(|line| {
                let line = line.expect("could not read line");
                let (station, temp) = line.split_once(';').expect("Error while splitting");
                let temp = onebrc::parse_temp(temp.as_bytes());
                let measurements_option = t_stations.get_mut(station);
                if let Some(measurements) = measurements_option {
                    measurements.update(temp);
                } else {
                    let measurements = onebrc::StationMeasurements {
                        min: temp,
                        max: temp,
                        count: 1,
                        sum: temp,
                    };
                    t_stations.insert(station.to_owned(), measurements);
                }
            });
            println!("Time reading lines in thread {i}={} μs", now_read_line.elapsed().as_micros());
            let now_insert_line = Instant::now();
            println!("Start inserting lines in thread {i}");
            for (station, measurements) in t_stations.iter() {
                let mut stations_guard = stations_clone.lock().expect("Error while locking");
                let joined_measurements_options = stations_guard.get_mut(station.as_str());
                if let Some(joined_measurements) = joined_measurements_options {
                    joined_measurements.merge(measurements);
                } else {
                    stations_guard.insert(station.to_owned(), *measurements);
                }
            }
            println!("Time inserting lines in thread {i}={} μs", now_insert_line.elapsed().as_micros());
        });
        handles.push(handle);
    }
    for handle in handles {
        handle.join().unwrap();
    }
    let mut stations: Vec<String> = stations.lock().unwrap().iter().map(|(&ref station, &ref measurements)| {
        let measurements = measurements.to_string();
        format!("{station}={measurements}")
    }).collect();
    stations.sort();
    let stations = stations.join(",");
    println!("{{{stations}}}");
    println!("Time={} μs", now.elapsed().as_micros());
}
